/*
 * Copyright (c) 2011-2022, The DART development contributors
 * All rights reserved.
 *
 * The list of contributors can be found at:
 *   https://github.com/dartsim/dart/blob/master/LICENSE
 *
 * This file is provided under the following "BSD-style" License:
 *   Redistribution and use in source and binary forms, with or
 *   without modification, are permitted provided that the following
 *   conditions are met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *   CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *   ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *   POSSIBILITY OF SUCH DAMAGE.
 */

#include "dart/common/Stopwatch.hpp"

namespace dart::common {

namespace {
StopwatchNS sw = StopwatchNS(true);
}

//==============================================================================
void tic()
{
  sw.reset();
}

//==============================================================================
double toc(bool print)
{
  return tocS(print);
}

//==============================================================================
double tocS(bool print)
{
  if (!print)
  {
    return sw.elapsedS();
  }

  sw.stop();
  const double elapsed = sw.elapsedS();
  std::cout << "Elapsed time is " << elapsed << " s." << std::endl;
  sw.start();
  return elapsed;
}

//==============================================================================
double tocMS(bool print)
{
  if (!print)
  {
    return sw.elapsedMS();
  }

  sw.stop();
  const double elapsed = sw.elapsedMS();
  std::cout << "Elapsed time is " << elapsed << " ms." << std::endl;
  sw.start();
  return elapsed;
}

//==============================================================================
double tocUS(bool print)
{
  if (!print)
  {
    return sw.elapsedUS();
  }

  sw.stop();
  const double elapsed = sw.elapsedUS();
  std::cout << "Elapsed time is " << elapsed << " us." << std::endl;
  sw.start();
  return elapsed;
}

//==============================================================================
double tocNS(bool print)
{
  if (!print)
  {
    return sw.elapsedNS();
  }

  sw.stop();
  const double elapsed = sw.elapsedNS();
  std::cout << "Elapsed time is " << elapsed << " ns." << std::endl;
  sw.start();
  return elapsed;
}

} // namespace dart::common
